/*
 * treefuncs.c
 *
 *
 * step-by-step add-on tree building procedures
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 *
 */

 
#include "puzzle.h"

 
/******************************************************************************/


/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void inittree()
{
	int i, maxbrnch;
	
	
	/* number of edges */
	maxbrnch = 2 * Maxspc - 3;
	
	/* allocate the memory for the whole tree */
	
	/* allocate memory for vector with all the edges of the tree */
	edge = (ONEEDGE *) calloc(maxbrnch, sizeof(ONEEDGE) );
	if (edge == NULL) maerror("edge [inittree]");

	/* allocate memory for vector with edge numbers of leaves */ 
	edgeofleaf = (int *) calloc(Maxspc, sizeof(int) );
	if (edgeofleaf == NULL) maerror("edgeofleaf [inittree]");
	
	/* allocate memory for all the edges the edge map */
	for (i = 0; i < maxbrnch; i++) {
		edge[i].edgemap = (int *) calloc(maxbrnch, sizeof(int) );
		if (edge[i].edgemap == NULL) maerror("edgemap [inittree]");
	}
		
	/* number all edges */
	for (i = 0; i < maxbrnch; i++) edge[i].numedge = i;
	
	/* initialize tree */
	
	nextedge = 3;
	nextleaf = 3;
	
	/* edge maps */
	(edge[0].edgemap)[0] = 0; /* you are on the right edge */
	(edge[0].edgemap)[1] = 4; /* go down left for leaf 1 */
	(edge[0].edgemap)[2] = 5; /* go down right for leaf 2 */
	(edge[1].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[1].edgemap)[1] = 0; /* you are on the right edge */
	(edge[1].edgemap)[2] = 3; /* go up/down right for leaf 2 */
	(edge[2].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[2].edgemap)[1] = 2; /* go up/down left for leaf 1 */
	(edge[2].edgemap)[2] = 0; /* you are on the right edge */
	
	/* interconnection */
	edge[0].up = NULL;
	edge[0].downleft = &edge[1];
	edge[0].downright = &edge[2];
	edge[1].up = &edge[0];
	edge[1].downleft = NULL;
	edge[1].downright = NULL;
	edge[2].up = &edge[0];
	edge[2].downleft = NULL;
	edge[2].downright = NULL;
	
	/* edges of leaves leaves */
	edgeofleaf[0] = 0;
	edgeofleaf[1] = 1;
	edgeofleaf[2] = 2;
}


/* add next leaf on the specified edge */
void addnextleaf(int dockedge)
{	
	int i;
	

	if (dockedge >= nextedge) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to nonexisting edge %d [addnextleaf]\n", dockedge);
		exit(1);
	}
	
	if (nextleaf >= Maxspc) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to a tree with %d leaves [addnextleaf]\n", Maxspc);
		exit(1);
	}
		
	/* necessary change in edgeofleaf if dockedge == edgeofleaf[0] */
	if (edgeofleaf[0] == dockedge) edgeofleaf[0] = nextedge;
	
	/* adding nextedge to the tree */
	edge[nextedge].up = edge[dockedge].up;
	edge[nextedge].downleft = &edge[dockedge];
	edge[nextedge].downright = &edge[nextedge+1];
	edge[dockedge].up = &edge[nextedge];

	if (edge[nextedge].up != NULL) {
		if ( ((edge[nextedge].up)->downleft) == &edge[dockedge] ) 
			(edge[nextedge].up)->downleft  = &edge[nextedge];
		else   
			(edge[nextedge].up)->downright = &edge[nextedge];
	}

	/* adding nextedge + 1 to the tree */
	edge[nextedge+1].up = &edge[nextedge];
	edge[nextedge+1].downleft = NULL;
	edge[nextedge+1].downright = NULL;
	edgeofleaf[nextleaf] = nextedge+1;
	
	/* the two new edges get info about the old edges */
	/* nextedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* down right changes to down left */
			case 5:		(edge[nextedge].edgemap)[i] = 4;
						break;
						
			/* null changes to down left */
			case 0:		(edge[nextedge].edgemap)[i] = 4;
						break;
			
			default: 	(edge[nextedge].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* nextedge + 1 */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down left changes to up */
			case 2:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
			
			/* up/down right changes to up */		
			case 3:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
						
			/* down left changes to up/down left */			
			case 4:		(edge[nextedge+1].edgemap)[i] = 2;
						break;

			/* down right changes to up/down left */	
			case 5:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* null changes to up/down left */
			case 0:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* up stays up */
			default: 	(edge[nextedge+1].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* dockedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down right changes to up */
			case 3:		(edge[dockedge].edgemap)[i] = 1;
						break;
						
			/* up/down left changes to up */
			case 2:		(edge[dockedge].edgemap)[i] = 1;
						break;
			
			default: 	break;
		}
	}

	/* all edgemaps are updated for the two new edges */
	/* nextedge */
	(edge[nextedge].edgemap)[nextedge] = 0;
	(edge[nextedge].edgemap)[nextedge+1] = 5; /* down right */
	
	/* nextedge + 1 */
	(edge[nextedge+1].edgemap)[nextedge] = 1; /* up */
	(edge[nextedge+1].edgemap)[nextedge+1] = 0;
	
	/* all other edges */
	for (i = 0; i < nextedge; i++) {
		(edge[i].edgemap)[nextedge] = (edge[i].edgemap)[dockedge];
		(edge[i].edgemap)[nextedge+1] = (edge[i].edgemap)[dockedge];
	}
	
	/* an extra for dockedge */
	(edge[dockedge].edgemap)[nextedge] = 1; /* up */
	(edge[dockedge].edgemap)[nextedge+1] = 3; /* up/down right */


	nextleaf++;
	nextedge = nextedge + 2;
}


/* free memory (to be called after inittree) */
void freetree()
{
	int i;

	for (i = 0; i < 2 * Maxspc - 3; i++) free(edge[i].edgemap);
	free(edge);
	free(edgeofleaf);	
}


/* show path from leaf A to B */
void showpath(int A, int B)
{	
	int curredge, finaledge, nextstep;
	
	
	fprintf(stderr, "\nEdges from leaf %d to leaf %d:  ", A, B);
	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	fprintf(stderr, "%d (%d)  ", curredge, edge[curredge].edgeinfo);
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;
				
			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;

			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;

			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;

		}
		fprintf(stderr, "%d (%d)  ", curredge, edge[curredge].edgeinfo);
	}	
}


/* show all current connections */
void showallconnections()
{
	int i, j;
	
	fprintf(stderr, "\n----- CONNECTIONS -----");
	
	for (i = 0; i < nextleaf; i++) {
		for (j = i + 1; j < nextleaf; j++) showpath(i, j);
	}
	
	fprintf(stderr, "\n");
}


/* clear all edgeinfos */
void resetedgeinfo()
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgeinfo = 0;
}


/* increment all edgeinfo between leaf A and B */
void incrementedgeinfo(int A, int B, int w)
{	
	int curredge, finaledge, nextstep;

	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	edge[curredge].edgeinfo = edge[curredge].edgeinfo + w;
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;
				
			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;
				
		}
		edge[curredge].edgeinfo = edge[curredge].edgeinfo + w;
	}	
}


/* prototype so that writeOTU knows about itself */
void writeOTU(FILE *outfp, int ed);


/* write OTU (recursive function) writes OTU sitting on edge ed */
void writeOTU(FILE *outfp, int ed)
{	
	int i;

	/* test whether we are on a leaf */
	if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
		for (i = 1; i < nextleaf; i++) {
			if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				column += fputid(outfp, trueID[i]);
				return;
			}
		}
	}

	/* we are NOT on a leaf */
	fprintf(outfp, "(");
	column++;
	writeOTU(outfp, edge[ed].downleft->numedge);
	fprintf(outfp, ", ");
	column++;
	column++;
	if (column > 55) {
		column = 2;
		fprintf(outfp, "\n  ");
	}
	writeOTU(outfp, edge[ed].downright->numedge);	
	fprintf(outfp, ")");
	column++;
}

 
/* write tree to treefile */
void writetree(FILE *outfp)
{	
	column = 1;
	fprintf(outfp, "(");
	column += fputid(outfp, trueID[0]) + 3;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downleft->numedge);
	column++;
	column++;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downright->numedge);	
	fprintf(outfp, ");\n");	
}


/* checks which edge has the lowest edgeinfo */
void minimumedgeinfo()
{
	int i;


	minedge = 0;
	mininfo = edge[0].edgeinfo;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgeinfo < mininfo) {
			minedge = i;
			mininfo = edge[i].edgeinfo;
		}
}


/* Look at the path from chooseA to chooseC and compare its edges to the edges found
   on the path from chooseA to chooseB. The first edge on A-C that is not also found
   on A-B is the cutting edge */
int findcuttingedge()
{
	int curredge, finaledgeB, nextstepB, finaledgeC, nextstepC;
	
	
	finaledgeB = edgeofleaf[chooseB]; /* final B edge */
	finaledgeC = edgeofleaf[chooseC]; /* final C edge */
	
	curredge = edgeofleaf[chooseA]; /* start edge */

	while (curredge != finaledgeC) {
		nextstepB = (edge[curredge].edgemap)[finaledgeB];
		nextstepC = (edge[curredge].edgemap)[finaledgeC];
		switch (nextstepC) {
		
			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;

			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;

			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;

		}
		if (nextstepB != nextstepC) return curredge;
	}
	
	fprintf(stderr, "\nERROR:  No cutting edge [findcuttingedge]\n");
	exit(1);
}


/* prototype for recursive function */
void searchpartition(int i);


/* analyse bipartition */
void searchpartition(int i)
{	
	int j;


	/* edge i belongs to cluster */
	edgelist[i] = TRUE;

	if ( edge[i].downright == NULL ||
		 edge[i].downleft == NULL) { /* if i is leaf */

		/* check out what leaf j sits on this edge i */		
		for (j = 1; j < nextleaf; j++) {
			if (edgeofleaf[j] == i) {
				leaflist[j] = TRUE; /* this leaf belongs to the cluster */
				return;
			}	
		}	
	} else { /* still on inner branch */
		searchpartition(edge[i].downleft->numedge);
		searchpartition(edge[i].downright->numedge);
	}
}


/* mark all edges in the cluster that is defined by cutting edge and
   that contains chooseA and chooseB as irrelvant for further analysis.
   declare in freeleaves only the leaf chooseA as free and
   reduce numleaves correspondingly */
void markcluster()
{	
	int i;


	/* check whether cutting edge is a internal branch or not */

	if ( edgeofleaf[0] == cuttingedge ||
		 edge[cuttingedge].downright == NULL ||
		 edge[cuttingedge].downleft == NULL) { /* external branch */

		/* game over - this is the add-on edge */
		numleaves = 2;
		for (i = 0; i < nextedge; i++) edge[i].edgeinfo = 1;
		edge[cuttingedge].edgeinfo = 0; /* minimum edge */

	} else { /* internal branch */

		/* initialise variables */
		for (i = 0; i < nextleaf; i++) leaflist[i] = FALSE;
		for (i = 0; i < nextedge; i++) edgelist[i] = FALSE;
	
		searchpartition(cuttingedge);
		
		/* check whether chooseA is TRUE or FALSE */
		if (leaflist[chooseA] == TRUE) { /* cluster ok */

			for (i = 0; i < nextleaf; i++) {
				if (leaflist[i] == TRUE) {
					if (freeleaves[i] == TRUE) {
						freeleaves[i] = FALSE;
						numleaves--;
					}
				}
			}
			for (i = 0; i < nextedge; i++) {
				if (edgelist[i] == TRUE) edge[i].edgeinfo = 1;
			}
			
		} else { /* cluster must be inverted */
			for (i = 0; i < nextleaf; i++) {
				if (leaflist[i] == FALSE) {
					if (freeleaves[i] == TRUE) {
						freeleaves[i] = FALSE;
						numleaves--;
					}
				}
			}
			for (i = 0; i < nextedge; i++) {
				if (edgelist[i] == FALSE) edge[i].edgeinfo = 1;
			}
			edge[cuttingedge].edgeinfo = 0;
		}
		
		/* chooseA remains active */
		freeleaves[chooseA] = TRUE;
		numleaves++;
	}
}
